//  Copyright © 2022 Optimove. All rights reserved.

import Foundation
import UserNotifications
import OptimoveCore

#if os(iOS) || os(watchOS) || os(tvOS)
    import UIKit
#endif

enum OSTypeID: NSNumber {
    case osTypeIDiOS = 1
    case osTypeIDOSX
    case osTypeIDAndroid
    case osTypeIDWindowsPhone
    case osTypeIDWindow
}

enum RuntimeType: NSNumber {
    case runtimeTypeUnknown = 0
    case runtimeTypeNative
    case runtimeTypeXamarin
    case runtimeTypeCordova
    case runtimeTypeJavaRuntime
}

enum TargetType: Int {
    case targetTypeDebug = 1
    case targetTypeRelease
}

struct Platform {
    static let isSimulator: Bool = {
        var isSim = false
        // if mac architechture and os is iOS, WatchOS or TVOS we're on a simulator
        #if targetEnvironment(simulator)
            isSim = true
        #endif
        return isSim
    }()
    
    static let isMacintosh: Bool = {
        var isMac = false
        // check architechture for mac
        #if (arch(i386) || arch(x86_64))
            isMac = true
        #endif
        return !isSimulator && isMac
    }()
}

extension Optimobile {
    
    func sendDeviceInformation() {
        
        var target = TargetType.targetTypeRelease
        
        //http://stackoverflow.com/questions/24111854/in-absence-of-preprocessor-macros-is-there-a-way-to-define-practical-scheme-spe
        #if DEBUG
            target = TargetType.targetTypeDebug
        #endif
        
        var app = [String : AnyObject]()
        app["bundle"] = Bundle.main.infoDictionary!["CFBundleIdentifier"] as AnyObject?
        app["version"] = Bundle.main.infoDictionary!["CFBundleShortVersionString"] as AnyObject?
        app["target"] = target.rawValue as AnyObject?
        
        
        var sdk = [String : AnyObject]()
        sdk["id"] = sdkType as AnyObject
        sdk["version"] = SDKVersion as AnyObject
        
        var runtime = [String : AnyObject]()
        var os = [String : AnyObject]()
        var device = [String : AnyObject]()
        
        runtime["id"] = RuntimeType.runtimeTypeNative.rawValue
        
        let timeZone = TimeZone.autoupdatingCurrent
        let tzName = timeZone.identifier
        device["tz"] = tzName as AnyObject?
        device["name"] = Sysctl.model as AnyObject?
        
        if Platform.isMacintosh {
            runtime["version"] = ProcessInfo.processInfo.operatingSystemVersionString as AnyObject?
            os["id"] = OSTypeID.osTypeIDOSX.rawValue
            os["version"] = ProcessInfo.processInfo.operatingSystemVersionString as AnyObject?
         }
        else {
            runtime["version"] = UIDevice.current.systemVersion as AnyObject?
            os["id"] = OSTypeID.osTypeIDiOS.rawValue
            os["version"] = UIDevice.current.systemVersion as AnyObject?
        }
        
        if (NSLocale.preferredLanguages.count >= 1) {
            device["locale"] = NSLocale.preferredLanguages[0] as AnyObject
        }
        
        device["isSimulator"] = Platform.isSimulator as AnyObject?
        
        let finalParameters = [
            "app" : app,
            "sdk" : sdk,
            "runtime" : runtime,
            "os" : os,
            "device" : device,
            "ios": self.getiOSAttrs()
        ]
        
        Optimobile.trackEvent(eventType: OptimobileEvent.STATS_CALL_HOME.rawValue, properties: finalParameters)
    }

    private func getiOSAttrs() -> [String:Any] {
        var push = [
            "scheduled": false,
            "timeSensitive": false
        ]

        if #available(iOS 15.0, *) {
            let permsBarrier = DispatchSemaphore(value: 0)

            UNUserNotificationCenter.current().getNotificationSettings { settings in
                push["scheduled"] = settings.scheduledDeliverySetting == .enabled
                push["timeSensitive"] = settings.timeSensitiveSetting == .enabled

                permsBarrier.signal()
            }

            _ = permsBarrier.wait(timeout: DispatchTime.now() + DispatchTimeInterval.seconds(5))
        }

        return [
            "hasGroup": AppGroupsHelper.isKumulosAppGroupDefined(),
            "push": push
        ]
    }
    
}

